"""
Formalisms to estimate incoming radiations when sky is not clouded.
"""
from datetime import datetime
from math import cos, exp, sin

from . import sun
from .extraterrestrial import incident_ground

corr_fac = dict(
    tropical=(0.95, 0.98, 1.02),
    midlatitude_summer=(0.97, 0.99, 1.02),
    subarctic_summer=(0.99, 0.99, 1.01),
    midlatitude_winter=(1.03, 1.01, 1.00)
)
"""(r0,r1,rk) correction factors

References: table2.8.1
"""


def atm_transmittance(cos_zenith, altitude, climate='midlatitude_summer'):
    """Transmittance of atmosphere for given sun position.

    References: eq2.8.1

    Args:
        cos_zenith (float): [-] cosine of sun zenith position
        altitude (float): [m] Altitude of observer (less than 2500)
        climate (str): type of climate see :obj:`corr_fac`

    Returns:
        (float): [-] transmittance factor
    """
    r0, r1, rk = corr_fac[climate]
    alt = altitude * 1e-3  # [m] to [km]

    a0 = r0 * (0.4237 - 0.00821 * (6 - alt) ** 2)
    a1 = r1 * (0.5055 + 0.00595 * (6.5 - alt) ** 2)
    k = rk * (0.2711 + 0.01858 * (2.5 - alt) ** 2)

    return a0 + a1 * exp(-k / cos_zenith)


def rg_ground(date, latitude, longitude=None, altitude=0., climate='midlatitude_summer'):
    """Incident energy reaching flat ground through atmosphere.

    Args:
        date (datetime): date and time of computation (solar time if longitude=None else UTC)
        latitude (float): [rad] Latitude north of the Equator
        longitude (float): [rad] Longitude positive East of Greenwich
        altitude (float): [m] Altitude of observer (less than 2500)
        climate (str): type of climate see :obj:`corr_fac`

    Returns:
        (float, float): [W.m-2] beam and diffuse energy reaching the ground on specific time
    """
    doy = date.timetuple().tm_yday
    declination = sun.declination(doy)
    time = sun.solar_hour_angle(date, longitude)

    rg_ext = incident_ground(doy, latitude, declination, time)
    if rg_ext <= 0.:
        return 0., 0.

    cos_zenith = cos(latitude) * cos(declination) * cos(time) + sin(latitude) * sin(declination)
    tau_b = atm_transmittance(cos_zenith, altitude, climate)

    rg_beam = rg_ext * tau_b
    rg_diff = rg_ext * (0.271 - 0.294 * tau_b)  # eq2.8.5

    return rg_beam, rg_diff



from duffie2013.diffuse import diff_frac


def test_diff_frac_is_between_0_and_1():
    rgs = (0, 1, 10, 20, 30, 40)
    for i, rg in enumerate(rgs):
        for rg_ext in rgs[i:]:
            assert 0 <= diff_frac(rg, rg_ext) <= 1



from datetime import datetime, timedelta
from math import pi, radians

import pytest
from duffie2013.extraterrestrial import (daily_incident_ground, incident_ground, incident_toa, rg_daily, rg_ground,
                                         rg_hourly)
from duffie2013.sun import declination, sunset_hour_angle


def test_incident_toa_varies_between_1320_and_1420():  # ref fig1.4.1
    for doy in range(1, 366):
        assert 1320 < incident_toa(doy, standard=True) < 1420
        assert 1320 < incident_toa(doy, standard=False) < 1420


def test_incident_ground_smaller_than_toa():
    for doy in range(1, 366, 30):
        decl = declination(doy)
        for latitude in (0, 10, 45):
            lat = radians(latitude)
            ws = sunset_hour_angle(lat, decl)
            for ind in range(-12, 12):
                assert 0 <= incident_ground(doy, lat, decl, ws * ind / 24) <= incident_toa(doy)


def test_incident_ground_sums_to_daily_incident_ground():
    for doy in range(1, 366, 30):
        decl = declination(doy)
        for latitude in (0, 10, 45):
            lat = radians(latitude)
            ws = sunset_hour_angle(lat, decl)
            rg_day = daily_incident_ground(doy, lat, decl, ws)

            sunrise = -ws * 12 / pi
            sunset = ws * 12 / pi
            rg_sum = sum(incident_ground(doy, lat, decl, h * pi / 12) * 3600e-6 if sunrise <= h <= sunset else 0
                         for h in range(-12, 12))

            assert rg_sum == pytest.approx(rg_day, abs=0.3)


def test_rg_daily_decreases_northward_in_winter():
    day = datetime(2020, 1, 1)
    rgs = [rg_daily(day, radians(latitude)) for latitude in (0, 15, 45, 50)]

    assert sorted(rgs, reverse=True) == rgs


def test_rg_daily_compares_to_table1_10_1():
    latitude = radians(45)
    for ind, rg_ref in enumerate([12.2, 17.4, 25.1, 33.2, 39.2, 41.7, 40.4, 35.3, 27.8, 19.6, 13.3, 10.7]):
        day = datetime(2020, ind + 1, 15)
        assert rg_daily(day, latitude) == pytest.approx(rg_ref, abs=0.5)


def test_rg_ground_sums_to_rg_daily():
    day = datetime(2020, 1, 1)
    latitude = radians(40)

    rg_day = rg_daily(day, latitude)
    rg_sum = sum(sum(rg_ground(day + timedelta(hours=h), latitude, longitude=0.)) * 3600e-6 for h in range(0, 24))

    assert rg_sum == pytest.approx(rg_day, abs=0.3)


def test_rg_hourly_equals_integration_of_rg_ground():
    date = datetime(2020, 6, 1, 10)
    latitude = radians(40)
    dt = timedelta(minutes=1)

    rg_h = sum(rg_hourly(date, latitude))
    rg_int = sum(sum(rg_ground(date + dt * i, latitude)) for i in range(60)) * dt.total_seconds() * 1e-6

    assert rg_h == pytest.approx(rg_int, abs=0.01)



"""
Maintain version for this package.
Do not edit this file, use 'version' section of config.
"""
# {# pkglts, version
#  -*- coding: utf-8 -*-

MAJOR = 0
"""(int) Version major component."""

MINOR = 1
"""(int) Version minor component."""

POST = 1
"""(int) Version post or bugfix component."""

__version__ = f"{MAJOR:d}.{MINOR:d}.{POST:d}"
# #}



"""
Formalisms to deal with the diffuse part of light.
"""


def diff_frac(rg, rg_ext):
    """Diffuse fraction of atmospheric radiation.

    References: eq2.10.1

    Args:
        rg (float): [W.m-2] or [MJ.m-2] measured radiation on flat surface on the ground
        rg_ext (float): [rg] theoretical extraterrestrial flat ground radiation

    Returns:
        (float): [-] fraction of diffuse in incoming radiation
    """
    if rg_ext < 1e-3:
        kt = 1.
    else:
        kt = rg / rg_ext

    if kt <= 0.2:
        return 1 - 0.09 * kt
    elif kt <= 0.8:
        return 0.9511 - 0.1604 * kt + 4.388 * kt ** 2 - 16.638 * kt ** 3 + 12.336 * kt ** 4
    else:
        return 0.165



"""
Amount of extraterrestrial radiation reaching the Earth
"""
from datetime import datetime
from math import cos, pi, sin

from .sun import declination as sun_declination, solar_constant, solar_hour_angle, sunset_hour_angle


def incident_toa(doy, standard=True):
    """Incident energy reaching top of atmosphere on given day.

    References: eq1.4.1

    Args:
        doy (int): day of year
        standard (bool): whether to use standard (default) or precise computation

    Returns:
        (float): [W.m-2]
    """
    if standard:
        e0 = (1 + 0.033 * cos(2 * pi * doy / 365))
    else:
        gamma = 2 * pi * (doy - 1) / 365
        e0 = (1.000110
              + 0.034221 * cos(gamma) + 0.001280 * sin(gamma)
              + 0.000719 * cos(2 * gamma) + 0.000077 * sin(2 * gamma))

    return solar_constant * e0


def incident_ground(doy, latitude, declination, time, standard=True):
    """Incident energy reaching flat ground not considering effects of atmosphere.

    References: eq1.10.2

    Args:
        doy (int): day of year
        latitude (float): [rad] latitude of ground
        declination (float): [rad] sun declination that day
        time (float): [rad] solar hour angle
        standard (bool): whether to use standard (default) or precise computation

    Returns:
        (float): [W.m-2] energy reaching the ground on specific time
    """
    proj = cos(latitude) * cos(declination) * cos(time) + sin(latitude) * sin(declination)

    return incident_toa(doy, standard) * proj


def daily_incident_ground(doy, latitude, declination, sunset, standard=True):
    """Daily amount of energy reaching flat ground not considering effects of atmosphere.

    Args:
        doy (int): day of year
        latitude (float): [rad] Latitude north of the Equator
        declination (float): [rad] sun declination that day
        sunset (float): [rad] solar hour angle of sunset that day
        standard (bool): whether to use standard (default) or precise computation

    Returns:
        (float): [MJ.m-2] energy reaching the ground integrated over the day
    """
    proj = cos(latitude) * cos(declination) * sin(sunset) + sunset * sin(latitude) * sin(declination)

    return 24 * 3600e-6 / pi * incident_toa(doy, standard) * proj


def rg_ground(date, latitude, longitude=None):
    """Incident energy reaching flat ground not considering effects of atmosphere.

    Notes: convenience API for :func:`incident_ground`

    Args:
        date (datetime): date and time of computation (solar time if longitude=None else UTC)
        latitude (float): [rad] Latitude north of the Equator
        longitude (float): [rad] Longitude positive East of Greenwich

    Returns:
        (float, float): [W.m-2] beam and diffuse energy reaching the ground on specific time
    """
    doy = date.timetuple().tm_yday
    rg = incident_ground(doy=doy,
                         latitude=latitude,
                         declination=sun_declination(doy),
                         time=solar_hour_angle(date, longitude))
    return max(0., rg), 0.


def rg_daily(day, latitude):
    """Daily amount of energy reaching flat ground not considering effects of atmosphere.

    Notes: convenience API for :func:`daily_incident_ground`

    Args:
        day (datetime): day of computation (the HH:MM:SS will be removed)
        latitude (float): [rad] Latitude north of the Equator

    Returns:
        (float): [MJ.m-2] total energy reaching the ground integrated over the day
    """
    doy = day.timetuple().tm_yday
    decl = sun_declination(doy)
    return daily_incident_ground(doy=doy,
                                 latitude=latitude,
                                 declination=decl,
                                 sunset=sunset_hour_angle(latitude, decl))


def rg_hourly(date, latitude, longitude=None):
    """Incident energy reaching flat ground not considering effects of atmosphere.

    Notes: integrate :func:`incident_ground` over one hour

    Args:
        date (datetime): date and time of computation (solar time if longitude=None else UTC)
        latitude (float): [rad] Latitude north of the Equator
        longitude (float): [rad] Longitude positive East of Greenwich

    Returns:
        (float, float): [MJ.m-2] hourly beam and diffuse energy reaching the
                        ground on specific time
    """
    doy = date.timetuple().tm_yday
    declination = sun_declination(doy)
    w1 = solar_hour_angle(date, longitude)
    w2 = w1 + pi / 12  # one hour later

    proj_int = 12 / pi * (cos(latitude) * cos(declination) * (sin(w2) - sin(w1))
                          + (w2 - w1) * sin(latitude) * sin(declination))

    return max(0., incident_toa(doy) * proj_int * 3600e-6), 0.



from datetime import datetime, timedelta
from math import acos, cos, pi, radians, sin, tan

solar_constant = 1367
"""[W.m-2] energy from the sun per unit time received on a unit area of surface
perpendicular to the direction of propagation of the radiation at mean earth-sun
distance outside the atmosphere

References: fig1.2.1
"""


def declination(doy, standard=True):
    """Sun declination.

    Angular position of the sun at solar noon (i.e., when the sun is on the
    local meridian) with respect to the plane of the equator.

    References: eq1.6.1a and eq1.6.1b

    Args:
        doy (int): day of year
        standard (bool): whether to use standard (default) or precise computation

    Returns:
        (float): [rad] sun declination that day
    """
    if standard:
        return radians(23.45) * sin(2 * pi * (284 + doy) / 365)
    else:
        gamma = 2 * pi * (doy - 1) / 365
        return (0.006918 - 0.399912 * cos(gamma) + 0.070257 * sin(gamma)
                - 0.006758 * cos(2 * gamma) + 0.000907 * sin(2 * gamma)
                - 0.002697 * cos(3 * gamma) + 0.00148 * sin(3 * gamma))


def sunrise_hour_angle(latitude, decl):
    """Hour angle of sunrise.

    References: eq1.6.10

    Args:
        latitude (float): [rad] latitude of place
        decl (float): [rad] sun declination angle

    Returns:
        (float): [rad] solar hour angle
    """
    return - sunset_hour_angle(latitude, decl)


def sunset_hour_angle(latitude, decl):
    """Hour angle of sunset.

    References: eq1.6.10

    Args:
        latitude (float): [rad] latitude of place
        decl (float): [rad] sun declination angle

    Returns:
        (float): [rad] solar hour angle
    """
    return acos(- tan(latitude) * tan(decl))


def day_length(latitude, decl):
    """Length of daylight.

    References: eq1.6.11

    Args:
        latitude (float): [rad] latitude of place
        decl (float): [rad] sun declination angle

    Returns:
        (float): [h] duration of sun light that day
    """
    return sunset_hour_angle(latitude, decl) * 24 / pi


def solar_time(date, longitude):
    """Convert date in UTC to local solar time.

    References: eq1.5.2 and eq1.5.3

    Args:
        date (datetime): UTC
        longitude (float): [rad] longitude of place (positive Eastward)

    Returns:
        (datetime): same instant but in solar time
    """
    doy = date.timetuple().tm_yday
    gamma = 2 * pi * (doy - 1) / 365  # eq1.4.2
    time_eq = 229.2 * (0.000075 + 0.001868 * cos(gamma) - 0.032077 * sin(gamma)
                       - 0.014615 * cos(2 * gamma) - 0.04089 * sin(2 * gamma))

    return date + timedelta(minutes=time_eq + longitude * 720 / pi)


def solar_hour(date, longitude=None):
    """Position of sun around the earth.

    solar_hour_angle expressed in hours instead of angle.

    Args:
        date (datetime): date and time of computation (solar time if longitude=None else UTC)
        longitude (float): [rad] Longitude positive East of Greenwich

    Returns:
        (float): [h] solar hour to solar noon
    """
    if longitude is not None:
        date = solar_time(date, longitude)

    return (date - date.replace(hour=12, minute=0, second=0, microsecond=0)).total_seconds() / 3600


def solar_hour_angle(date, longitude=None):
    """Position of sun around the earth.

    Observing the sun from earth, the solar hour angle is an expression of time,
    expressed in angular measurement, usually degrees, from solar noon. At solar
    noon the hour angle is zero degrees, with the time before solar noon expressed
    as negative degrees, and the local time after solar noon expressed as positive
    degrees.

    For example, at 10:30 AM local apparent time the hour angle is −22.5°
    (15° per hour times 1.5 hours before noon).

    Args:
        date (datetime): date and time of computation (solar time if longitude=None else UTC)
        longitude (float): [rad] Longitude positive East of Greenwich

    Returns:
        (float): [rad] solar hour angle
    """
    return solar_hour(date, longitude) * pi / 12



from datetime import datetime
from math import pi, radians

import pytest
from duffie2013.sun import day_length, declination, solar_time, sunrise_hour_angle, sunset_hour_angle


def test_declination_stays_between_both_tropics():  # fig1.6.3
    cancer = radians(23.4365)
    capricorn = radians(-23.4365)

    for doy in range(1, 366, 30):
        assert capricorn <= declination(doy, True) <= cancer
        assert capricorn <= declination(doy, False) <= cancer


def test_sunset_hour_angle_works_correctly():  # fig1.6.3
    sunset_ref = (19 + 50 / 60 - 12) * pi / 12
    assert sunset_hour_angle(radians(50), radians(21)) == pytest.approx(sunset_ref, abs=0.01)
    sunset_ref = (16 + 10 / 60 - 12) * pi / 12
    assert sunset_hour_angle(radians(50), -radians(21)) == pytest.approx(sunset_ref, abs=0.01)


def test_sunrise_hour_angle_is_before_sunset_hour_angle():
    assert sunrise_hour_angle(radians(50), radians(21)) < sunset_hour_angle(radians(50), radians(21))


def test_day_length_works_correctly():  # fig1.6.3
    assert day_length(radians(50), radians(21)) == pytest.approx(15.7, abs=0.1)
    assert day_length(radians(50), -radians(21)) == pytest.approx(8.3, abs=0.1)


def test_solar_time_equals_utc_when_longitude_equal_0():
    for date in (datetime(2020, 6, 1, 12),
                 datetime(1978, 3, 1, 10),
                 datetime(2000, 9, 12, 20),
                 datetime(2019, 12, 31, 23, 59)):
        assert abs(solar_time(date, 0.) - date).total_seconds() / 60 < 16  # fig1.5.1


def test_solar_time_increases_eastward():
    for date in (datetime(2020, 6, 1, 12),
                 datetime(1978, 3, 1, 10),
                 datetime(2000, 9, 12, 20),
                 datetime(2019, 12, 31, 23, 59)):
        sts = [solar_time(date, radians(longitude)) for longitude in (-45, -30, -15, 0, 10, 50)]

        assert sorted(sts) == sts



# {# pkglts, pysetup.kwds
# format setup arguments
from pathlib import Path
from setuptools import setup, find_packages

short_descr = "Implementation of formalisms from Duffie 2013 on solar engineering"
readme = open('README.rst').read()
history = open('HISTORY.rst').read()

# find packages
pkgs = find_packages('src')

src_dir = Path("src/duffie2013")

data_files = []
for pth in src_dir.rglob("*"):
    if not pth.is_dir() and "__pycache__" not in pth.parts:
        if pth.suffix in ['.json', '.ini', '.csv', '.rst', '.svg']:
            data_files.append(str(pth.relative_to(src_dir)))

pkg_data = {'duffie2013': data_files}

setup_kwds = dict(
    name='duffie2013',
    version="0.1.1",
    description=short_descr,
    long_description=readme + '\n\n' + history,
    author="revesansparole",
    author_email="revesansparole@gmail.com",
    url='https://gitlab.com/revesansparole/duffie2013',
    license='cc_by_nc',
    zip_safe=False,

    packages=pkgs,
    
    package_dir={'': 'src'},
    
    
    package_data=pkg_data,
    setup_requires=[
        "pytest-runner",
        ],
    install_requires=[
        ],
    tests_require=[
        "coverage",
        "pytest",
        "pytest-cov",
        "pytest-mock",
        ],
    entry_points={},
    keywords='',
    
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    )
# #}
# change setup_kwds below before the next pkglts tag

# do not change things below
# {# pkglts, pysetup.call
setup(**setup_kwds)
# #}



from datetime import datetime, timedelta
from math import cos, radians

import pytest
from duffie2013.clear_sky import atm_transmittance, rg_ground


def test_atm_transmittance_is_between_0_and_1():
    for zen in (0, 30, 45, 60, 90):
        for alt in (0, 100, 400, 1000, 2500):
            assert 0 <= atm_transmittance(cos(radians(zen)), alt) <= 1


def test_atm_transmittance_increases_with_altitude():
    cos_zenith = cos(radians(45))
    trs = [atm_transmittance(cos_zenith, alt) for alt in (0, 100, 400, 1000, 2500)]

    assert sorted(trs) == trs


def test_atm_transmittance_decreases_with_zenith_angle():
    trs = [atm_transmittance(cos(radians(zen)), 100) for zen in (0, 30, 45, 60, 90)]

    assert sorted(trs, reverse=True) == trs


def test_rg_ground_compares_to_ex2_8_1():
    rgb, rgd = rg_ground(datetime(2020, 8, 21, 11, 30), radians(43), altitude=270)

    assert rgb == pytest.approx(702, abs=1)
    assert rgd == pytest.approx(101, abs=1)


def test_rg_ground_compares_to_ex2_8_2():
    latitude = radians(43)
    altitude = 270
    day = datetime(2020, 8, 21)  # why not 22? don't know

    for h, rgd_ref, rg_ref in [(11.5, 0.36, 2.89),
                               (10.5, 0.35, 2.69),
                               (9.5, 0.34, 2.31),
                               (8.5, 0.32, 1.78),
                               (7.5, 0.28, 1.15),
                               (6.5, 0.20, 0.53),
                               (5.5, 0.05, 0.07),
                               ]:
        rgb, rgd = rg_ground(day + timedelta(hours=h), latitude, altitude)

        assert rgd * 3600e-6 == pytest.approx(rgd_ref, abs=1)
        assert (rgb + rgd) * 3600e-6 == pytest.approx(rg_ref, abs=1)
